(* ::Package:: *)

BeginPackage["SemiAlgInv`"];


InvS::usage="InvS[S,f,H] LZZ decision procedure"
DiffInv::usage="DiffInv[S,f,H] Differential invariant check"
NSSBC::usage="NSSBC[S,f,H] Non-smooth strict barrier certificate"
Lf::usage="Checks for membership of the vector field inside the contingent cone of a set given by Min[Max[g1,...,gn]...]."
VCGen::usage="Generates verification conditions for non-smooth barrier certificates of the form Min[Max[g1,...,gn]...]."


Begin["`Private`"]


(* INVARIANT CHECKING SECTION *)


DiffInd[e_,t_] :=
 Module[{fml},
fml = LogicalExpand[e];
If[FreeQ[fml,_Not|_ForAll|_Some],
(* during differential induction, all literals need to be sustained differentially, hence replace \[Or] by \[And] right away *)
(fml = fml /.Or->And;
If[FreeQ[fml,Not[_]|_Or],
(fml/.
(* weaken > to >= and < to <= during DiffInd *)
{eq_Greater :>GreaterEqual@@eq,eq_Less :>LessEqual@@eq})
/. (* map D over left and right part of (in)equalities *){eq_Equal|eq_GreaterEqual|eq_LessEqual|eq_Greater|eq_Less :>Map[D[#,t]&,eq],
eq_Unequal :> (Message[DiffInd::illegal,eq]; $Failed)
},
(* Or not yet implemented. Not is rejected *)
(Message[DiffInd::illegal,fml];$Failed)
]),
(* Or not yet implemented. Not is rejected *)
(Message[DiffInd::illegal,fml];$Failed)
]
]
DiffInd[e_,t_,ODE_]:= (DiffInd[e,t]/.Flatten[Map[ToRules,ODE]])


IDiffInd[e_] :=
 Module[{fml},
fml = LogicalExpand[e] ;
If[FreeQ[fml,_Not|_ForAll|_Some],
(* during differential induction, all literals need to be sustained differentially, hence replace \[Or] by \[And] right away *)
(fml=fml/.Or->And;
If[FreeQ[fml,Not[_]|_Or],
(fml/.
(* weaken > to >= and < to <= during DiffInd *)
{eq_Greater :>GreaterEqual@@eq,eq_Less :>LessEqual@@eq})
/. (* map Dt over left and right part of (in)equalities *){eq_Equal|eq_GreaterEqual|eq_LessEqual|eq_Greater|eq_Less :>Map[Dt[#]&,eq],
eq_Unequal :> (Message[IDiffInd::illegal,eq]; $Failed)
},
(* Or not yet implemented. Not is rejected *)
(Message[IDiffInd::illegal,fml];$Failed)
]
),
(* Or not yet implemented. Not is rejected *)
(Message[IDiffInd::illegal,fml];$Failed)
]
]
IDiffInd[e_,f_, vars_]:=((IDiffInd[e]/.Flatten[Map[ToRules, (Apply[Equal,Transpose[{Map[Derivative[1][#]&, vars],f}],1])/.Derivative[1][x_]:>Dt[x]]]) (* any remaining differentials are 0 as not in ODE *)/._Dt->0)


DiffInv[Inv_,f_,vars_,H_]:=Module[{cond=ForAll[vars, Implies[H, IDiffInd[Inv,f,vars]]]},
Reduce[cond, Reals]
]


(* Set righ-hand side of terms to zero *)
ZeroRHS[formula_] := Module[{},formula/.{
Equal[a_,b_]        :>  Equal[a-b,0],
Unequal[a_,b_]      :>  Unequal[a-b,0],
Greater[a_,b_]      :>  Greater[a-b,0],
GreaterEqual[a_,b_] :>  GreaterEqual[a-b,0],
Less[a_,b_]         :>  Less[a-b,0], 
LessEqual[a_,b_]    :>  LessEqual[a-b,0]
}]


GeqToLeq[formula_]:=Module[{}, formula/.{         GreaterEqual[lhs_,rhs_] :>  LessEqual[rhs,lhs]} ] 
GtToLt[formula_]:=Module[{}, formula/.{           Greater[lhs_,rhs_]      :>  Less[rhs,lhs]} ] 
UnequalToLtOrGt[formula_]:=Module[{}, formula/.{  Unequal[lhs_,rhs_]      :>  Or[Less[lhs,rhs] ,Less[rhs,lhs]]} ] 
EqualToLeqAndGeq[formula_]:=Module[{}, formula/.{ Equal[lhs_,rhs_]        :>  And[LessEqual[lhs,rhs] ,LessEqual[rhs,lhs]]} ] 
LeqToLtOrEqual[formula_]:=Module[{}, formula/.{   LessEqual[lhs_,rhs_]    :>  Or[Less[lhs,rhs] ,Equal[rhs,lhs]]} ] 


PreProcess[expression_]:=Module[{},BooleanMinimize[ZeroRHS[ GeqToLeq[ GtToLt[ UnequalToLtOrGt[BooleanMinimize[ expression, "DNF"] ] ] ] ], "DNF"]]


PreProcessNSSBC[expression_]:=Module[{},BooleanMinimize[ZeroRHS[ GeqToLeq[ GtToLt[  UnequalToLtOrGt[BooleanMinimize[ expression, "DNF"] ] ] ] ], "DNF"]]


(* Lie derivative *)
LD[p_, f_List, vars_List]:=Grad[p,vars].f


(* Smoothness condition: non-vanishing gradient *)
NonZeroGrad[p_, f_List, vars_List]:=LogicalExpand[Grad[p,vars]!= Map[Function[x,0],vars]]


(* Given a polynomial, compute a list of its Lie derivatives up to order n *)
NLieDerivatives[p_,n_, f_List, vars_List]:=Module[{},
NestList[LD[#,f,vars]&,p,n]
]


(* Given a polynomial, compute the condition requiring that its first n-1 Lie derivatives are 0 and its nth Lie derivative is < 0 *)
NthLieLtZero[p_,n_,f_List,vars_List]:=Module[{},
If[n==0, p<0,
If[n==1, p==0 && LD[p,f,vars]<0,
NLie = NLieDerivatives[p,n-1,f,vars];
NthLieCondition = LD[Last[NLie],f,vars]<0;
Apply[And,Map[Function[x,x==0],NLie]] && NthLieCondition
]]
]


(* Given a polynomial, compute the condition requiring that its first n-1 Lie derivatives are 0 and its nth Lie derivative is <= 0 *)
NthLieLeqZero[p_,n_,f_List,vars_List]:=Module[{},
If[n==0, p<0,
If[n==1, p==0 && LD[p,f,vars]<=0,
NLie = NLieDerivatives[p,n-1,f,vars];
NthLieCondition = LD[Last[NLie],f,vars]<=0;
Apply[And,Map[Function[x,x==0],NLie]] && NthLieCondition
]]
]


(* Given a polynomial, compute the condition requiring that its first n Lie derivatives are all 0 *)
NthLieEqZero[p_,n_,f_List,vars_List]:=Module[{},
NLie = NLieDerivatives[p,n,f,vars];
Apply[And,Map[Function[x,x==0],NLie]] 
]


(* Given a polynomial p and an integer n, check that the ideal <p, Lfp, Lf^2p, ..., Lf^np > is saturated under adding Lie derivatives of higher order  *)
Rank[p_,n_,f_List,vars_List]:=Module[{},
NLie=NLieDerivatives[p,n,f,vars];
NPlusOneLie = LD[Last[NLie],f,vars];
GB = GroebnerBasis[NLie, vars, MonomialOrder -> DegreeReverseLexicographic];
Remainder = PolynomialReduce[NPlusOneLie, GB, vars, MonomialOrder -> DegreeReverseLexicographic][[2]]
]


(* Given a polynomial p, compute its rank Rk(p), i.e. the smallest order n of the Lie derivative that saturates the ideal <p, Lfp, Lf^2p, ... > *)
(* buggy: Rank[x,{x},{x}] gives 1 instead of 0 *)
(*
Rank[p_,f_List,vars_List]:=Module[{},
rem=1;
n=0;
While[Not[PossibleZeroQ[rem]],
n++;
rem=Rank[p,n,f,vars];
];
n
]
*)


(* Given a polynomial p, compute its rank Rk(p), i.e. the smallest order n of the Lie derivative that saturates the ideal <p, Lfp, Lf^2p, ... > *)
Rank[p_,f_List,vars_List]:=Module[{n=0},
While[Not[PossibleZeroQ[Rank[p,n,f,vars]]],
n++;
];
n
]


InfpStrict[p_, f_List, vars_List]:=Module[{rank = Rank[p,f,vars]},
countToRank =Table[i,{i,0,rank}];
Apply[Or, Map[Function[x,NthLieLtZero[p,x,f,vars]], countToRank] ]
]


InfpNonStrict[p_, f_List, vars_List]:=Module[{rank = Rank[p,f,vars]},
countToRank =Table[i,{i,0,rank-1}];
Apply[Or, Map[Function[x,NthLieLtZero[p, x,f,vars]], countToRank] ] || NthLieLeqZero[p, rank,f,vars]
]


InfpEqual[p_, f_List, vars_List]:=Module[{rank = Rank[p,f,vars]}, NthLieEqZero[p, rank,f,vars] ]


InfS[S_, f_List, vars_List]:=Module[{processedS=PreProcess[S]},
processedS/.{
LessEqual[p_,0]:> InfpNonStrict[p, f,vars], 
Equal[p_,0]:> InfpEqual[p, f,vars], 
Less[p_,0]:>InfpStrict[p, f,vars]}
]


IvInfS[S_, f_List, vars_List]:=Module[{processedS=PreProcess[S]},
processedS/.{
LessEqual[p_,0] :> InfpNonStrict[p, -f,vars],
Equal[p_,0] :> InfpEqual[p, -f,vars], 
Less[p_,0] :>InfpStrict[p, -f,vars]}
]


ComplementS[S_]:=Module[{},PreProcess[Not[S]] ]


InvS[S_, f_List, vars_List, H_]:=Module[{
Cond2 = Implies[S && H && InfS[H,f,vars], InfS[S,f,vars]],
Cond3 = Implies[ComplementS[S] && H && IvInfS[H,f,vars], ComplementS[IvInfS[S,f,vars]]]
},
Resolve[ForAll[vars, Cond2 && Cond3], Reals] 
]


(* Differentiable case *)
Lf[p_, f_List, vars_List]:=Grad[p,vars].f<0


(* Conjunctive case - max functions *)
Lf[max[p_], f_List, vars_List]:=Lf[p,f,vars]

Lf[max[p1_,pn__], f_List, vars_List]:=And[
Implies[p1>max[pn],Lf[p1,f,vars]],
Implies[max[pn]>p1,Lf[max[pn],f,vars]],
Implies[p1==max[pn],And[Lf[p1,f,vars],Lf[max[pn],f,vars]]]
]


(* Disjunctive case - min functions *)
Lf[min[p_], f_List, vars_List]:=Lf[p,f,vars]

Lf[min[p1_,pn__], f_List, vars_List]:=And[
Implies[p1<min[pn],Lf[p1,f,vars]],
Implies[min[pn]<p1,Lf[min[pn],f,vars]],
Implies[p1==min[pn],Or[Lf[p1,f,vars],Lf[min[pn],f,vars]]]
]


toMinMaxForm[formula_]:=Module[{},
ZeroRHS[LogicalExpand[formula]/.{GreaterEqual[a_,b_]:> LessEqual[b,a]}]/.{And :> max, Or:>  min}/.{LessEqual[a_,0]:> a}]


NSSBC[formula_, f_List, vars_List, H_]:=Module[{
precessedform=PreProcessNSSBC[formula],
minmax=toMinMaxForm[formula]},
rightform=FreeQ[precessedform,Equal]&&FreeQ[precessedform,Greater]&&FreeQ[precessedform,Less]&&FreeQ[precessedform,Unequal] && FreeQ[precessedform,Equal];
cond=If[rightform,
(* Reduce[ForAll[vars,Implies[H && minmax==0/.{max :>Max, min:> Min},Lf[minmax,f,vars]/.{max :> Max, min:> Min}]],Reals],*)
Reduce[Implies[H && minmax==0/.{max :>Max, min:> Min},Lf[minmax,f,vars]/.{max :> Max, min:> Min}],vars,Reals],
False];
Reduce[ForAll[vars,cond],Reals]
]


End[]
EndPackage[]
